@@ -12,49 +12,63 @@
 #
 ##############################################################################
 import gc
-import platform
 import unittest
 
 from persistent.interfaces import UPTODATE
+from persistent._compat import PYPY
+from persistent.tests.utils import skipIfNoCExtension
+
 
 # pylint:disable=protected-access,too-many-lines,too-many-public-methods
-
-_is_pypy = platform.python_implementation() == 'PyPy'
+# pylint:disable=attribute-defined-outside-init,redefined-outer-name
 
 _marker = object()
 
-def skipIfNoCExtension(o):
-    import persistent
-    return unittest.skipIf(
-        persistent._cPickleCache is None,
-        "The C extension is not available")(o)
+class DummyPersistent(object):
+    _Persistent__ring = None
 
-def skipIfPurePython(o):
-    import persistent._compat
-    return unittest.skipIf(
-        persistent._compat.PURE_PYTHON,
-        "Cannot mix and match implementations"
-    )(o)
+    def _p_invalidate(self):
+        from persistent.interfaces import GHOST
+        self._p_state = GHOST
 
-class PickleCacheTests(unittest.TestCase):
+    _p_deactivate = _p_invalidate
+
+    def _p_invalidate_deactivate_helper(self, clear=True):
+        self._p_invalidate()
+
+    def _p_activate(self):
+        self._p_state = UPTODATE
+
+
+class DummyConnection(object):
+    pass
+
+
+class ClosedConnection(DummyConnection):
+    def __init__(self, test):
+        self.test = test
+
+    def setstate(self, obj): # pragma: no cover
+        self.test.fail("Connection is closed")
+
+    def register(self, obj):
+        """Does nothing."""
+
+def _len(seq):
+    return len(list(seq))
+
+
+class PickleCacheTestMixin(object):
 
     # py2/3 compat
     assertRaisesRegex = getattr(unittest.TestCase,
                                 'assertRaisesRegex',
                                 unittest.TestCase.assertRaisesRegexp)
 
-
-    def setUp(self):
-        import persistent.picklecache
-        self.orig_types = persistent.picklecache._CACHEABLE_TYPES
-        persistent.picklecache._CACHEABLE_TYPES += (DummyPersistent,)
-
-    def tearDown(self):
-        import persistent.picklecache
-        persistent.picklecache._CACHEABLE_TYPES = self.orig_types
-
     def _getTargetClass(self):
-        from persistent.picklecache import PickleCache
+        from persistent.picklecache import PickleCachePy as BasePickleCache
+        class PickleCache(BasePickleCache):
+            _CACHEABLE_TYPES = BasePickleCache._CACHEABLE_TYPES + (DummyPersistent,)
         return PickleCache
 
     def _getTargetInterface(self):
@@ -69,6 +83,10 @@ class PickleCacheTests(unittest.TestCase):
     def _getDummyPersistentClass(self):
         return DummyPersistent
 
+    def _getRealPersistentClass(self):
+        from persistent.persistence import PersistentPy
+        return PersistentPy
+
     def _makePersist(self, state=None, oid=b'foo', jar=_marker, kind=_marker):
         from persistent.interfaces import GHOST
 
@@ -122,7 +140,7 @@ class PickleCacheTests(unittest.TestCase):
         cache = self._makeOne()
         default = object
 
-        self.assertTrue(cache.get('nonesuch', default) is default)
+        self.assertIs(cache.get('nonesuch', default), default)
 
     def test___setitem___non_string_oid_raises_TypeError(self):
         cache = self._makeOne()
@@ -333,10 +351,10 @@ class PickleCacheTests(unittest.TestCase):
         self.assertEqual(items[9][0], b'oid_0099')
 
         for oid in oids[:90]:
-            self.assertTrue(cache.get(oid) is None)
+            self.assertIsNone(cache.get(oid))
 
         for oid in oids[90:]:
-            self.assertFalse(cache.get(oid) is None)
+            self.assertIsNotNone(cache.get(oid))
 
     def test_incrgc_w_smaller_drain_resistance(self):
         cache = self._makeOne()
@@ -373,9 +391,86 @@ class PickleCacheTests(unittest.TestCase):
         for oid in oids:
             self.assertTrue(cache.get(oid) is None)
 
+    def test_full_sweep_clears_weakrefs_in_interface(self, sweep_method='full_sweep'):
+        # https://github.com/zopefoundation/persistent/issues/149
+        # Sweeping the cache clears weak refs (for PyPy especially)
+        # In the real world, this shows up in the interaction with
+        # persistent objects and zope.interface/zope.component,
+        # so we use an Interface to demonstrate. This helps find issues
+        # like needing to run GC more than once, etc, because of how the
+        # object is referenced.
+        from zope.interface import Interface
+
+        gc.disable()
+        self.addCleanup(gc.enable)
+
+        jar = ClosedConnection(self)
+        cache = self._makeOne(jar, 0)
+
+        # Make a persistent object, put it in the cache as saved
+        class P(self._getRealPersistentClass()):
+            "A real persistent object that can be weak referenced"
+
+        p = P()
+        p._p_jar = jar
+        p._p_oid = b'\x01' * 8
+        cache[p._p_oid] = p
+        p._p_changed = False
+
+        # Now, take a weak reference to it from somewhere far away.
+        Interface.subscribe(p)
+
+        # Remove the original object
+        del p
+
+        # Sweep the cache.
+        getattr(cache, sweep_method)()
+
+        # Now, try to use that weak reference. If the weak reference is
+        # still around, this will raise the error about the connection
+        # being closed.
+        Interface.changed(None)
+
+    def test_incrgc_clears_weakrefs_in_interface(self):
+        self.test_full_sweep_clears_weakrefs_in_interface(sweep_method='incrgc')
+
+    def test_full_sweep_clears_weakrefs(self, sweep_method='incrgc'):
+        # like test_full_sweep_clears_weakrefs_in_interface,
+        # but directly using a weakref. This is the simplest version of the test.
+        from weakref import ref as WeakRef
+        gc.disable()
+        self.addCleanup(gc.enable)
+
+        jar = ClosedConnection(self)
+        cache = self._makeOne(jar, 0)
+
+        # Make a persistent object, put it in the cache as saved
+        class P(self._getRealPersistentClass()):
+            """A real persistent object that can be weak referenced."""
+
+        p = P()
+        p._p_jar = jar
+        p._p_oid = b'\x01' * 8
+        cache[p._p_oid] = p
+        p._p_changed = False
+
+        # Now, take a weak reference to it
+        ref = WeakRef(p)
+
+        # Remove the original object
+        del p
+
+        # Sweep the cache.
+        getattr(cache, sweep_method)()
+
+        # Now, try to use that weak reference; it should be gone.
+        p = ref()
+        self.assertIsNone(p)
+
+    def test_incrgc_clears_weakrefs(self):
+        self.test_full_sweep_clears_weakrefs(sweep_method='incrgc')
 
     def test_minimize(self):
-
         cache = self._makeOne()
         oids = self._populate_cache(cache)
         self.assertEqual(cache.cache_non_ghost_count, 100)
@@ -644,8 +739,19 @@ class PickleCacheTests(unittest.TestCase):
 
         self.assertTrue(pclass.invalidated)
 
+    def test_cache_raw(self):
+        raw = self._makePersist(kind=self._getRealPersistentClass())
+        cache = self._makeOne(raw._p_jar)
 
-class PythonPickleCacheTests(PickleCacheTests):
+        cache[raw._p_oid] = raw
+        self.assertEqual(1, len(cache))
+        self.assertIs(cache.get(raw._p_oid), raw)
+
+        del raw
+        self.assertEqual(1, len(cache))
+
+
+class PythonPickleCacheTests(PickleCacheTestMixin, unittest.TestCase):
     # Tests that depend on the implementation details of the
     # Python PickleCache and the Python persistent object.
     # Anything that involves directly setting the _p_state of a persistent
@@ -669,13 +775,9 @@ class PythonPickleCacheTests(PickleCacheTests):
 
         p._p_deactivate = bad_deactivate
 
-        import persistent.picklecache
-        sweep_types = persistent.picklecache._SWEEPABLE_TYPES
-        persistent.picklecache._SWEEPABLE_TYPES = DummyPersistent
-        try:
-            self.assertEqual(cache.full_sweep(), 0)
-        finally:
-            persistent.picklecache._SWEEPABLE_TYPES = sweep_types
+        cache._SWEEPABLE_TYPES = DummyPersistent
+        self.assertEqual(cache.full_sweep(), 0)
+        del cache._SWEEPABLE_TYPES
 
         del p._p_deactivate
         self.assertEqual(cache.full_sweep(), 1)
@@ -750,7 +852,6 @@ class PythonPickleCacheTests(PickleCacheTests):
         self.assertEqual(Pclass._p_jar, cache.jar)
         self.assertIs(cache.persistent_classes[key], Pclass)
 
-
     def test_mru_nonesuch_raises_KeyError(self):
         cache = self._makeOne()
 
@@ -986,8 +1087,7 @@ class PythonPickleCacheTests(PickleCacheTests):
         self.assertTrue(items[0][1] is candidate)
         self.assertEqual(candidate._p_state, UPTODATE)
 
-    def test_cache_garbage_collection_bytes_also_deactivates_object(self,
-                                                                    force_collect=_is_pypy):
+    def test_cache_garbage_collection_bytes_also_deactivates_object(self):
 
         class MyPersistent(self._getDummyPersistentClass()):
             def _p_deactivate(self):
@@ -1029,16 +1129,16 @@ class PythonPickleCacheTests(PickleCacheTests):
         cache.incrgc()
         self.assertEqual(cache.total_estimated_size, 0)
 
-        # It also shrank the measured size of the cache;
-        # this would fail under PyPy if _SWEEP_NEEDS_GC was False
-        if force_collect: # pragma: no cover
+        # It also shrank the measured size of the cache,
+        # though this may require a GC to be visible.
+        if PYPY: # pragma: no cover
             gc.collect()
         self.assertEqual(len(cache), 1)
 
 
     def test_new_ghost_obj_already_in_cache(self):
-        cache, key, candidate = super(PythonPickleCacheTests, self).test_new_ghost_obj_already_in_cache()
-
+        base_result = super(PythonPickleCacheTests, self).test_new_ghost_obj_already_in_cache()
+        cache, key, candidate = base_result
         # If we're sneaky and remove the OID and jar, then we get the duplicate
         # key error. Removing them only works because we're not using a real
         # persistent object.
@@ -1090,21 +1190,75 @@ class PythonPickleCacheTests(PickleCacheTests):
         self.assertEqual(cache.cache_non_ghost_count, 0)
         self.assertEqual(len(cache), 0)
 
+    def test_interpreter_finalization_ffi_cleanup(self):
+        # When the interpreter is busy garbage collecting old objects
+        # and clearing their __dict__ in random orders, the CFFI cleanup
+        # ``ffi.gc()`` cleanup hooks we use on CPython don't
+        # raise errors.
+        #
+        # Prior to Python 3.8, when ``sys.unraisablehook`` was added,
+        # the only way to know if this test fails is to look for AttributeError
+        # on stderr.
+        #
+        # But wait, it gets worse. Prior to https://foss.heptapod.net/pypy/cffi/-/issues/492
+        # (CFFI > 1.14.5, unreleased at this writing), CFFI ignores
+        # ``sys.unraisablehook``, so even on 3.8 the only way to know
+        # a failure is to watch stderr.
+        #
+        # See https://github.com/zopefoundation/persistent/issues/150
+
+        import sys
+        unraised = []
+        try:
+            old_hook = sys.unraisablehook
+        except AttributeError:
+            pass
+        else: # pragma: no cover
+            sys.unraisablehook = unraised.append
+            self.addCleanup(setattr, sys, 'unraisablehook', old_hook)
+
+        cache = self._makeOne()
+        oid = self._numbered_oid(42)
+        o = cache[oid] = self._makePersist(oid=oid)
+        # Clear the dict, or at least part of it.
+        # This is coupled to ``cleanup_hook``
+        if cache.data.cleanup_hook:
+            del cache.data._addr_to_oid
+        del cache[oid]
+
+        self.assertEqual(unraised, [])
+
 
 @skipIfNoCExtension
-@skipIfPurePython
-class CPickleCacheTests(PickleCacheTests):
+class CPickleCacheTests(PickleCacheTestMixin, unittest.TestCase):
 
     def _getTargetClass(self):
-        from persistent.cPickleCache import PickleCache
-        return PickleCache
+        from persistent._compat import _c_optimizations_available as get_c
+        return get_c()['persistent.picklecache'].PickleCache
+
+    def _getRealPersistentClass(self):
+        from persistent._compat import _c_optimizations_available as get_c
+        return get_c()['persistent.persistence'].Persistent
 
     def _getDummyPersistentClass(self):
-        from persistent import Persistent
-        class DummyPersistent(Persistent):
+        class DummyPersistent(self._getRealPersistentClass()):
             __slots__ = ()
         return DummyPersistent
 
+    def test_inst_does_not_conform_to_IExtendedPickleCache(self):
+        # Test that ``@use_c_impl`` is only applying the correct
+        # interface declaration to the C implementation.
+        from persistent.interfaces import IExtendedPickleCache
+        from zope.interface.verify import verifyObject
+        from zope.interface.exceptions import Invalid
+        # We don't claim to implement it.
+        self.assertFalse(IExtendedPickleCache.providedBy(self._makeOne()))
+        # And we don't even provide everything it asks for.
+        # (Exact error depends on version of zope.interface and what we
+        # fail to implement. 5.0 probably raises MultipleInvalid).
+        with self.assertRaises(Invalid):
+            verifyObject(IExtendedPickleCache, self._makeOne(), tentative=True)
+
     def test___setitem___persistent_class(self):
         cache = super(CPickleCacheTests, self).test___setitem___persistent_class()
         self.assertEqual(_len(cache.items()), 1)
@@ -1112,17 +1266,17 @@ class CPickleCacheTests(PickleCacheTests):
     def test_cache_garbage_collection_bytes_with_cache_size_0(self):
 
         class DummyConnection(object):
-          def register(self, obj):
-            pass
+            def register(self, obj):
+                pass
 
         dummy_connection = DummyConnection()
         dummy_connection.register(1) # for coveralls
 
         def makePersistent(oid):
-           persist = self._getDummyPersistentClass()()
-           persist._p_oid = oid
-           persist._p_jar = dummy_connection
-           return persist
+            persist = self._getDummyPersistentClass()()
+            persist._p_oid = oid
+            persist._p_jar = dummy_connection
+            return persist
 
         cache = self._getTargetClass()(dummy_connection)
         dummy_connection._cache = cache
@@ -1154,25 +1308,29 @@ class CPickleCacheTests(PickleCacheTests):
         self.assertEqual(len(cache), 0)
 
 
-class DummyPersistent(object):
+class TestWeakValueDictionary(unittest.TestCase):
 
-    def _p_invalidate(self):
-        from persistent.interfaces import GHOST
-        self._p_state = GHOST
+    def _getTargetClass(self):
+        from persistent.picklecache import _WeakValueDictionary
+        return _WeakValueDictionary
 
-    _p_deactivate = _p_invalidate
+    def _makeOne(self):
+        return self._getTargetClass()()
 
-    def _p_invalidate_deactivate_helper(self, clear=True):
-        self._p_invalidate()
+    @unittest.skipIf(PYPY, "PyPy doesn't have the cleanup_hook")
+    def test_cleanup_hook_gc(self):
+        # A more targeted test than ``test_interpreter_finalization_ffi_cleanup``
+        # See https://github.com/zopefoundation/persistent/issues/150
+        wvd = self._makeOne()
 
-    def _p_activate(self):
-        self._p_state = UPTODATE
+        class cdata(object):
+            o = object()
+            pobj_id = id(o)
+        wvd['key'] = cdata.o
 
-class DummyConnection(object):
-    pass
+        wvd.__dict__.clear()
+        wvd.cleanup_hook(cdata)
 
-def _len(seq):
-    return len(list(seq))
 
 def test_suite():
     return unittest.defaultTestLoader.loadTestsFromName(__name__)
